/*
 *  Flatcurve KWin decoration
 *  Based on RedHat's Bluecurve
 *
 *  Copyright (C) 1999, 2001 Daniel Duley <mosfet@kde.org>
 *  Matthias Ettrich <ettrich@kde.org>
 *  Karol Szwed <gallium@kde.org>
 *
 *  Draws mini titlebars for tool windows.
 *  Many features are now customizable.
 */

#ifndef FLATCURVE_H
#define FLATCURVE_H

#include <qbutton.h>
#include <qbitmap.h>
#include <kpixmap.h>
#include <kwin/client.h>
#include <kwin/kwinbutton.h>


class QSpacerItem;
class QHBoxLayout;

namespace KWinInternal {

class FlatcurveHandler: public QObject
{
	public:
		FlatcurveHandler();
		~FlatcurveHandler();
		void reset();

	private:
  void readConfig();
		void createPixmaps();
		void freePixmaps();
		void drawButtonBackground(KPixmap *pix, 
				const QColorGroup &g, bool sunken);
  void recolor( QImage &img, const QColor& color );
};

enum ButtonPos { ButtonLeft = 0, ButtonMid, ButtonRight };

class FlatcurveButton : public KWinInternal::KWinButton
{
	public:
		FlatcurveButton(Client *parent=0, const char *name=0, 
			 bool largeButton=true, int pos=ButtonMid,
			 bool isStickyButton=false, const unsigned char *bitmap=NULL,
			 const QString& tip=NULL);
		~FlatcurveButton(); 

		int last_button;
		void turnOn( bool isOn );
		void setBitmap(const unsigned char *bitmap);
		QSize sizeHint() const;
                int     pos;

	protected:
  void resizeEvent( QResizeEvent* e);
  void showEvent(QShowEvent *ev);
  void doShape();
		void enterEvent(QEvent *);
		void leaveEvent(QEvent *);
		void mousePressEvent( QMouseEvent* e );
		void mouseReleaseEvent( QMouseEvent* e );
		void drawButton(QPainter *p);
		void drawButtonLabel(QPainter*) {;}

		QBitmap* deco;
		bool    large;
		bool    isLeft;
		bool    isSticky;
		bool	isMouseOver;
		Client* client;
};


class FlatcurveClient : public KWinInternal::Client
{
	Q_OBJECT

	public:
		FlatcurveClient( Workspace *ws, WId w, QWidget *parent=0, 
						  const char *name=0 );
		~FlatcurveClient() {;}

	protected:
		void resizeEvent( QResizeEvent* );
		void paintEvent( QPaintEvent* );
		void showEvent( QShowEvent* );
		void mouseDoubleClickEvent( QMouseEvent * );
		void captionChange( const QString& name );
		void maximizeChange(bool m);
		void activeChange(bool);
		void iconChange();
		void stickyChange(bool on);
		MousePosition mousePosition(const QPoint &) const;

	protected slots:
		void slotMaximize();
		void menuButtonPressed();

	private:
		void doShape();
		void calcHiddenButtons();
		void addClientButtons( const QString& s, bool isLeft=true );

		enum Buttons{ BtnHelp=0, BtnMax, BtnIconify, BtnClose, 
					  BtnMenu, BtnSticky, BtnCount };
  FlatcurveButton* button[ FlatcurveClient::BtnCount ];
		int           lastButtonWidth;
		int           titleHeight;
		bool          largeButtons;
		QHBoxLayout*  hb;
		QSpacerItem*  titlebar;
};

};

#endif
// vim: ts=4
